/*
 * JORAM: Java(TM) Open Reliable Asynchronous Messaging
 * Copyright (C) 2004 - 2010 ScalAgent Distributed Technologies
 * Copyright (C) 1996 - 2000 Dyade
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * Initial developer(s): ScalAgent Distributed Technologies
 * Contributor(s):
 */
package org.objectweb.joram.mom.dest;

import java.util.Properties;

import org.objectweb.joram.mom.notifications.AbstractRequestNot;
import org.objectweb.joram.mom.notifications.ClientMessages;
import org.objectweb.joram.mom.notifications.ExceptionReply;
import org.objectweb.joram.mom.notifications.FwdAdminRequestNot;
import org.objectweb.joram.mom.notifications.GetRightsRequestNot;
import org.objectweb.joram.mom.notifications.RequestGroupNot;
import org.objectweb.joram.mom.notifications.SpecialAdminRequest;
import org.objectweb.joram.mom.notifications.WakeUpNot;
import org.objectweb.joram.shared.excepts.MomException;
import org.objectweb.joram.shared.excepts.RequestException;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import fr.dyade.aaa.agent.Agent;
import fr.dyade.aaa.agent.AgentId;
import fr.dyade.aaa.agent.AgentServer;
import fr.dyade.aaa.agent.Channel;
import fr.dyade.aaa.agent.DeleteNot;
import fr.dyade.aaa.agent.Notification;
import fr.dyade.aaa.agent.UnknownAgent;
import fr.dyade.aaa.agent.UnknownNotificationException;
import fr.dyade.aaa.agent.WakeUpTask;
import fr.dyade.aaa.common.Debug;
import fr.dyade.aaa.util.management.MXWrapper;

/**
 * A <code>Destination</code> agent is an agent hosting a MOM destination,
 * for example a <tt>Queue</tt> or a <tt>Topic</tt>.
 * Its behavior is provided by a <code>DestinationImpl</code> instance.
 *
 * @see AdminDestinationItf
 */
public abstract class Destination extends Agent implements AdminDestinationItf {
  /** logger */
  public static Logger logger = Debug.getLogger(Destination.class.getName());
  
  /**
   * The reference of the <code>DestinationItf</code> instance providing this
   * this agent with its <tt>Destination</tt> behaviour.
   */
  protected DestinationImpl destImpl;

  public void setFreeReading(boolean freeReading) {
    destImpl.freeReading = freeReading;
  }
  
  public void setFreeWriting(boolean freeWriting) {
    destImpl.freeWriting = freeWriting;
  }
  
  protected transient WakeUpTask task;
  
  /**
   * Empty constructor for newInstance(). 
   */ 
  public Destination() {}

  /**
   *  Constructor with parameters for fixing the destination and specifying its
   * identifier.
   *  It is uniquely used by the AdminTopic agent.
   */
  protected Destination(String name, boolean fixed, int stamp) {
    super(name, fixed, stamp);
  }

  /**
   * Initializes the destination by creating the <tt>DestinationItf</tt>
   * object.
   *
   * @param adminId  Identifier of the destination administrator.
   * @param props    The initial set of properties.
   */
  public final void init(AgentId adminId, Properties props) throws RequestException {
    destImpl = createsImpl(adminId, props);
    destImpl.setAgent(this);
  }

  /**
   * Creates the specific implementation.
   *
   * @param adminId  Identifier of the topic administrator.
   * @param prop     The initial set of properties.
   * @throws RequestException 
   */
  public abstract DestinationImpl createsImpl(AgentId adminId, Properties prop) throws RequestException;

  /**
   * Returns the type of this destination: Queue or Topic.
   * 
   * @return the type of this destination.
   * 
   * @see org.objectweb.joram.shared.DestinationConstants#TOPIC_TYPE
   * @see org.objectweb.joram.shared.DestinationConstants#QUEUE_TYPE
   */
  public abstract byte getType();
  
  /**
   * Gives this agent an opportunity to initialize after having been deployed,
   * and each time it is loaded into memory.
   *
   * @param firstTime		true when first called by the factory
   *
   * @exception Exception
   *	unspecialized exception
   */
  protected void agentInitialize(boolean firstTime) throws Exception {
    if (logger.isLoggable(BasicLevel.DEBUG))
      logger.log(BasicLevel.DEBUG, "agentInitialize(" + firstTime + ')');

    super.agentInitialize(firstTime);
    destImpl.setAgent(this);
    destImpl.initialize(firstTime);
    
    if (destImpl.getPeriod() > -1)
      task = new WakeUpTask(getId(), WakeUpNot.class, destImpl.getPeriod());
    
    try {
      MXWrapper.registerMBean(destImpl, getMBeanName());
    } catch (Exception exc) {
      logger.log(BasicLevel.ERROR, this + " jmx failed", exc);
    }
  }

  /** Finalizes the agent before it is garbaged. */
  public void agentFinalize(boolean lastTime) {
    if (task != null)
      task.cancel();
    try {
      MXWrapper.unregisterMBean(getMBeanName());
    } catch (Exception exc) {
      if (logger.isLoggable(BasicLevel.DEBUG))
        logger.log(BasicLevel.DEBUG, "Destination.agentFinalize", exc);
    }
    super.agentFinalize(lastTime);
  }

  public String getMBeanName() {
    StringBuffer strbuf = new StringBuffer();
    
    strbuf.append("Joram#").append(AgentServer.getServerId());
    strbuf.append(':');
    strbuf.append("type=Destination,name=").append(getName());
    
    return strbuf.toString();
  }

  /**
   * Distributes the received notifications to the appropriate reactions.
   * @throws Exception 
   */
  public void react(AgentId from, Notification not) throws Exception {
    if (logger.isLoggable(BasicLevel.DEBUG))
      logger.log(BasicLevel.DEBUG,
                 "Destination.react(" + from + ',' + not + ')');

    // set agent no save (this is the default).
    setNoSave();
    
    try {
      if (not instanceof GetRightsRequestNot)
        destImpl.getRights(from, (GetRightsRequestNot) not);
      else if (not instanceof SpecialAdminRequest)
        destImpl.specialAdminRequest(from, (SpecialAdminRequest) not);
      else if (not instanceof ClientMessages)
        destImpl.clientMessages(from, (ClientMessages) not);
      else if (not instanceof UnknownAgent)
        destImpl.unknownAgent(from, (UnknownAgent) not);
      else if (not instanceof RequestGroupNot)
        destImpl.requestGroupNot(from, (RequestGroupNot)not);
      else if (not instanceof DeleteNot) {
        destImpl.deleteNot(from, (DeleteNot) not); 
        if (destImpl.canBeDeleted()) {
          // A DeleteNot notification is finally processed at the
          // Agent level when its processing went successful in
          // the DestinationItf instance.
          super.react(from, not);
        }
      } else if (not instanceof WakeUpNot) {
        setNoSave();
        if (task == null || ((WakeUpNot) not).update) {
          setPeriod(destImpl.getPeriod());
        }
        if (destImpl.getPeriod() > 0) {
          destImpl.wakeUpNot((WakeUpNot) not);
        }
      } else if (not instanceof FwdAdminRequestNot)
        destImpl.handleAdminRequestNot(from, (FwdAdminRequestNot) not);
      else
        throw new UnknownNotificationException(not.getClass().getName());
    } catch (MomException exc) {
      // MOM Exceptions are sent to the requester.
      if (logger.isLoggable(BasicLevel.WARN))
        logger.log(BasicLevel.WARN, this + ".react()", exc);

      AbstractRequestNot req = (AbstractRequestNot) not;
      Channel.sendTo(from, new ExceptionReply(req, exc));
    } catch (UnknownNotificationException exc) {
      super.react(from, not);
    }
  }
  
  private void setPeriod(long period) {
    if (logger.isLoggable(BasicLevel.DEBUG))
      logger.log(BasicLevel.DEBUG, this + ": setPeriod(" + period + ")." + " -> task " + task);
    if (task == null) {
      task = new WakeUpTask(getId(), WakeUpNot.class, period);
    } else {
      // cancel task
      task.cancel();
      // Schedules the wake up task period.
      if (period > 0)
        task = new WakeUpTask(getId(), WakeUpNot.class, period);
    }
  }
  
  protected void setNoSave() {
    if (logger.isLoggable(BasicLevel.DEBUG))
      logger.log(BasicLevel.DEBUG, this + ": setNoSave().");
    super.setNoSave();
  }

  protected void setSave() {
    if (logger.isLoggable(BasicLevel.DEBUG))
      logger.log(BasicLevel.DEBUG, this + ": setSave().");
    super.setSave();
  }
}
